#!/usr/bin/perl

# Article:
# [Pereira and Jamhour, 2011] 
# Metodo Heuristico para Rotular Grupos em Sistema de Deteccao de Intrusao baseado em Anomalia

# URL: http://ids.celepar.pr.gov.br/scripts

# Normalize requests

use strict;

my $file = $ARGV[0] || die "Example: perl $0 dts.p00.txt\nResults: dts.p00.txt.dsn\n";
my $file_dsn = $file.".dsn";

my @ds = &load_dataset($file);
# &print_ds(@ds);
my @dsn = &normalize_dataset(@ds);
# &print_dsn(\@ds,\@dsn);
&save_dataset($file_dsn,@dsn);
print "done - $file_dsn\n";

###############################################################################
# Load HTTP requests
# $ds[$i][0] = source address
# $ds[$i][1] = label (anomaly, attack or normal)
# $ds[$i][2] = resource path
# $ds[$i][3] = resource query
# $ds[$i][4] = host
# $ds[$i][5] = user-agent
# $ds[$i][6] = cookie
# $ds[$i][7] = referer
# $ds[$i][8] = content
###############################################################################
sub load_dataset {
   my $file = shift;
   my $count = -1;
   my $pointer = 0;
   my @ds;
   my @fields; 
   my $line;
   open (FILE, $file);
   while ($line = <FILE>) {
      chomp($line);
      if ($line eq "") {
         $pointer = 0;
      }
      if ($pointer == 3) {
         @fields = ($line =~ m/(.*?)(\\r\\n|\\n)/);
         $ds[$count][8] .= $fields[0];                                  # content
      }
      if ($pointer == 2) {
         if ($line =~ /^(\\r)?\\n$/) {
            $pointer = 3;
         } else {
            @fields = ($line =~ m/(.*?)\:(.*?)(\\r\\n|\\n)/);
            if ($fields[0] =~ /^Host$/i) {
               $ds[$count][4] = $fields[1];                             # host
            } elsif ($fields[0] =~ /^User-Agent$/i) {
               $ds[$count][5] = $fields[1];                             # user-agent
            } elsif ($fields[0] =~ /^Cookie$/i) {
               $ds[$count][6] = $fields[1];                             # cookie
            } elsif ($fields[0] =~ /^Referer$/i) {
               $ds[$count][7] = $fields[1];                             # referer
            }
         }
      }
      if ($pointer == 1) {
         @fields = split(/ /,$line);
         ($ds[$count][2]) = split(/\?/,$fields[1]);                     # resource - path
         $ds[$count][3] = substr($fields[1],length($ds[$count][2]));    # resource - query
         $pointer = 2;
      }
      if ($line =~ /^REQ /) {
         $count++;
         $pointer = 1;
         @fields = split(/ +/,$line);
         $ds[$count][0] = $fields[2];                                   # source address
         $ds[$count][1] = $fields[7];                                   # label
         if ((($count+1) % 10000) == 0) {
            # view
            print "load - item(".($count+1).")\n";
         }
      }
   }
   close (FILE);
   return @ds;
}
###############################################################################


###############################################################################
# Print dataset
sub print_ds {
   my @ds = @_;
   my ($i,$j);
   for ($i = 0; $i < scalar(@ds); $i++) {
      for ($j = 0; $j <= 8; $j++) {
         print $j." ".$ds[$i][$j]."\n";
      }
      print "\n\n";
   }
}
###############################################################################


###############################################################################
# Normalize HTTP requests
# $dsn[$i][2..8][0] = Quantity of alphabetical [a-zA-Z]
# $dsn[$i][2..8][1] = Quantity of numerical [0-9]
# $dsn[$i][2..8][2] = Quantity of non-alphanumerical (other) [^a-zA-Z0-9]
###############################################################################
sub normalize_dataset {
   my @ds = @_;
   my @dsn;
   my $alp = 0; # alpha
   my $num = 0; # numerical
   my $oth = 0; # other - non-alphanum
   my ($i,$j);
   for ($i = 0; $i < scalar(@ds); $i++) {
      $dsn[$i][0] = $ds[$i][0];
      $dsn[$i][1] = $ds[$i][1];
      for ($j = 2; $j <= 8; $j++) {
        $alp = $ds[$i][$j];
        $num = $ds[$i][$j];
        $oth = $ds[$i][$j];
        $alp =~ s/[^a-z]//ig;
        $num =~ s/[^0-9]//g;
        $oth =~ s/[a-z0-9]//ig;
        $dsn[$i][$j][0] = length($alp);
        $dsn[$i][$j][1] = length($num);
        $dsn[$i][$j][2] = length($oth);
      }
      if ((($i+1) % 10000) == 0) {
         # view
         print "normalize - item(".($i+1).")\n";
      }
   }
   return @dsn;
}
###############################################################################


###############################################################################
# Print normalized dataset
sub print_dsn {
   my ($ref_ds,$ref_dsn) = @_;
   my @ds = @{$ref_ds};
   my @dsn = @{$ref_dsn};
   for (my $i = 0; $i < scalar(@ds); $i++) {
      print "0 ".$ds[$i][0]."\n"; 
      print "1 ".$ds[$i][1]."\n"; 
      for (my $j = 2; $j <= 8; $j++) {
         print $j." ".$ds[$i][$j]."\n";
         print "  alpha(".$dsn[$i][$j][0].") num(".$dsn[$i][$j][1].") other(".$dsn[$i][$j][2].")\n";
      }
      print "\n\n";
   }
}
###############################################################################


###############################################################################
sub save_dataset {
   my $file = shift;
   my @dsn = @_;
   open (FILE, "> $file");
   print FILE "src_addr label up:alpha num other uq:alpha num other ho:alpha num other ua:alpha num other ck:alpha num other rf:alpha num other ct:alpha num other\n";
   for (my $i = 0; $i < scalar(@dsn); $i++) {
      print FILE $dsn[$i][0]." "; # source address
      print FILE $dsn[$i][1]." "; # label
      for (my $j = 2; $j <= 8; $j++) {
         print FILE $dsn[$i][$j][0]." ".$dsn[$i][$j][1]." ".$dsn[$i][$j][2]." ";
      }
      if ((($i+1) % 10000) == 0) {
         # view
         print "write - item(".($i+1).")\n";
      }
      print FILE "\n";
   }
   close (FILE);
}
###############################################################################


