#!/usr/bin/perl

# Article:
# [Pereira and Jamhour, 2011]
# Metodo Heuristico para Rotular Grupos em Sistema de Deteccao de Intrusao baseado em Anomalia

# URL: http://ids.celepar.pr.gov.br/scripts

# Clustering

use strict;

my $file = $ARGV[0] || die "Example: perl $0 dts.p00.txt.dsn 100\nResults: dataset.txt.dsn.100.cls, dataset.txt.dsn.100.dsc and dataset.txt.dsn.100.dbi\n";
my $width = $ARGV[1] || 100;
my $file_dsn = $file;
my $file_cls = $file.".".$width.".cls";
my $file_dsc = $file.".".$width.".dsc";
my $file_dbi = $file.".".$width.".dbi";

my @dsn = load_dsn($file_dsn);
my ($ref_cls,$ref_dsc) = &clustering($width,@dsn);

my @cls = @{$ref_cls};
my @dsc = @{$ref_dsc};

&save_clusters($file_cls,@cls);
print "done - $file_cls\n";
&save_dsclusters($file_dsc,@dsc);
print "done - $file_dsc\n";
my $dbi = &daviesbouldin(@cls);
&save_dbi($file_dbi,$dbi);
print "done - $file_dbi\n";


###############################################################################
sub load_dsn {
   my $file = shift;
   my $fline = 1;
   my @dsn;
   my $c = 0;
   my $line;
   open (FILE, $file);
   while ($line = <FILE>) {
      if (! $fline) {
         chomp ($line);
         @{$dsn[$c]} = split(/ /,$line);
         $c++;
         if ((($c+1) % 10000) == 0) {
            # view
            print "load - item(".($c+1).")\n";
         }
      } else {
         $fline = 0;
      }
   }
   close (FILE);
   return @dsn;
}
###############################################################################


###############################################################################
sub clustering {
   my $width = shift;
   my @dsc = @_;
   my @cls;
   my $iter = 0;
   my ($i,$j,$k,$l,$min_k,$min_d);
   my ($euc_sum,$distance);
   while ($iter <= 1) {
      if ($iter > 0) {
         for ($k = 0; $k < scalar(@cls); $k++) {
            $cls[$k][0] = 0;
            $cls[$k][1] = 0;
         }
      }
      for ($i = 0; $i < scalar(@dsc); $i++) {
         $min_k = -1;
         $min_d = -1;
         for ($k = 0; $k < scalar(@cls); $k++) {
            $distance = 0;
            for ($l = 2; $l <= 20; $l += 3) {
               $euc_sum = 0;
               for ($j = $l; $j < $l + 3; $j++) {
                  $euc_sum += (($dsc[$i][$j] - $cls[$k][$j]) ** 2);
               }
               $distance += sqrt($euc_sum);
            }
            if ($distance <= $width && ($distance < $min_d || $min_k == -1)) {
               $min_k = $k;
               $min_d = $distance;
            }
         }
         if ($min_k == -1) { # new cluster
            $k = scalar(@cls);
            $cls[$k][0] = 1;
            $cls[$k][1] = 0;
            for ($j = 2; $j <= 22; $j++) {
               $cls[$k][$j] = $dsc[$i][$j];
            } 
         } else {
            $dsc[$i][1] = $min_k;
            $cls[$min_k][0]++; 
            $cls[$min_k][1] += $min_d;   # use: davies-bouldin 
         }
         if ((($i+1) % 1000) == 0) { 
            # view
            $k = scalar(@cls);
            print "width($width) iter($iter) clusters($k) item(".($i+1).")\n";
         }
      }
      $iter++;
   }
   return (\@cls,\@dsc);
}
###############################################################################


###############################################################################
# Davies-bouldin - Clustering Validity Index
sub daviesbouldin {
   my @cls = @_;
   my $sum = 0;
   my $count = 0;
   my ($a,$b,$j,$l,$max,$euc_sum,$distance,$diam_a,$diam_b,$calc);
   for ($a = 0; $a < scalar(@cls); $a++) {
      $max = 0;
      for ($b = 0; $b < scalar(@cls); $b++) {
         if ($a != $b) {
            $distance = 0;
            for ($l = 2; $l <= 20; $l += 3) {
               $euc_sum = 0;
               for ($j = $l; $j < $l + 3; $j++) {
                  $euc_sum += (($cls[$a][$j] - $cls[$b][$j]) ** 2);
               }
               $distance += sqrt($euc_sum);
            }
            $diam_a = $cls[$a][1] / $cls[$a][0];
            $diam_b = $cls[$b][1] / $cls[$b][0];
            $calc = ($diam_a + $diam_b) / $distance;
            if ($calc > $max) {
               $max = $calc;
            }
         }
      }
      $sum += $max;
      $count++;
   }
   my $db = $sum / $count;
   return $db;
}
###############################################################################


###############################################################################
# Save clusters
sub save_clusters {
   my $file = shift;
   my @cls = @_;
   my ($i,$j);
   open (FILE, "> $file");
   print FILE "id requests sum_dists up:alpha num other uq:alpha num other ho:alpha num other ua:alpha num other ck:alpha num other rf:alpha num other ct:alpha num other\n";
   for ($i = 0; $i < scalar(@cls); $i++) { 
      print FILE $i." ";
      for ($j = 0; $j < scalar(@{$cls[$i]}); $j++) {
         print FILE $cls[$i][$j]." ";
      }
      print FILE "\n";
   }
   close (FILE);
}
###############################################################################


###############################################################################
# Save dataset labeled with clusters
sub save_dsclusters {
   my $file = shift;
   my @dsc = @_;
   my $i;
   open (FILE, "> $file");
   print FILE "cluster_id source_addr\n";
   for ($i = 0; $i < scalar(@dsc); $i++) { 
      print FILE $dsc[$i][1]." ";
      print FILE $dsc[$i][0]." ";
      print FILE "\n";
   }
   close (FILE);
}
###############################################################################


###############################################################################
# Save DBI 
sub save_dbi {
   my $file = shift;
   my $dbi = shift;
   open (FILE, "> $file");
   print FILE "Davies-bouldin index: $dbi\n";
   close (FILE);
}
###############################################################################
