#!/usr/bin/perl

# Article:
# [Pereira and Jamhour, 2011]
# Metodo Heuristico para Rotular Grupos em Sistema de Deteccao de Intrusao baseado em Anomalia

# URL: http://ids.celepar.pr.gov.br/scripts

# Intrusion Detection

# D1 pXX labeled in pYY dataset (Portnoy)
# D2 pYY labeled in pYY dataset (Zhong)
# D3 pXX labeled in pYY dataset with $decision 

use strict;

my $ifdie = "Example(D1): perl $0 dts.p00.txt.dsn.100.cls.portnoy.0.15.lbl dts.p01.txt.dsn 0\nResults: dts.p00.txt.dsn.100.cls.portnoy.0.15.lbl.in.dts.p01.txt.dsn.0.dtc,\n dts.p00.txt.dsn.100.cls.portnoy.0.15.lbl.in.dts.p01.txt.dsn.0.res\n\n";
$ifdie .= "Example(D2): perl $0 dts.p01.txt.dsn.100.cls.portnoy.0.15.lbl dts.p01.txt.dsn 0\nResults: dts.p01.txt.dsn.100.cls.portnoy.0.15.lbl.in.dts.p01.txt.dsn.0.dtc,\n dts.p01.txt.dsn.100.cls.portnoy.0.15.lbl.in.dts.p01.txt.dsn.0.res\n\n";
$ifdie .= "Example(D3): perl $0 dts.p00.txt.dsn.100.cls.portnoy.0.15.lbl dts.p01.txt.dsn 100\nResults: dts.p00.txt.dsn.100.cls.portnoy.0.15.lbl.in.dts.p01.txt.dsn.100.dtc,\n dts.p00.txt.dsn.100.cls.portnoy.0.15.lbl.in.dts.p01.txt.dsn.100.res\n\n";
my $file = $ARGV[0] || die "$ifdie";
my $file2 = $ARGV[1] || die "$ifdie";
my $decision = $ARGV[2] || 0; # 0  = closest cluster (D1 or D2)
                              # >0 = closest cluster with width (D3)

my $file_lbl = $file;
my $file_dsn = $file2;
my $file_dtc = $file.".in.".$file2.".".$decision.".dtc";
my $file_res = $file.".in.".$file2.".".$decision.".res";

my @lbl = &load_labeledclusters($file_lbl);
my @dsn = &load_dsn($file_dsn);

my @dtc = &detecting ($decision,\@lbl,\@dsn);
&save_dtc($file_dtc,@dtc);
print "done - $file_dtc\n";

my @res = &detection_results(@dtc);
&save_res($file_res,@res);
print "done - $file_res\n";


##############################################################################
sub load_labeledclusters {
   my $file = shift;
   my @lbl;
   my ($k,$line);
   my $fline = 1;
   open (FILE, $file);
   while ($line = <FILE>) {
      if (! $fline) {
         chomp($line);
         @{$lbl[$k]} = split(/ +/,$line);
         $k++;
      } else {
         $fline = 0;
      }
   }
   close (FILE);
   return @lbl;
}
##############################################################################


###############################################################################
sub load_dsn {
   my $file = shift;
   my $fline = 1;
   my @dsn;
   my $c = 0;
   my $line;
   open (FILE, $file);
   while ($line = <FILE>) {
      if (! $fline) {
         chomp ($line);
         @{$dsn[$c]} = split(/ /,$line);
         $c++;
         if ((($c+1) % 10000) == 0) {
            # view
            print "load - item(".($c+1).")\n";
         }
      } else {
         $fline = 0;
      }
   }
   close (FILE);
   return @dsn;
}
###############################################################################


###############################################################################
sub detecting {
   my $decision = shift;
   my ($ref_lbl,$ref_dsn) = @_;
   my @lbl = @{$ref_lbl};
   my @dsn = @{$ref_dsn};
   my @dtc;
   my ($i,$j,$k,$l,$min_k,$min_d,$lbl_i,$distance,$euc_sum);
   for (my $i = 0; $i < scalar(@dsn); $i++) {
      $min_k = -1;
      $min_d = -1;
      $lbl_i = $dsn[$i][1]; 
      for ($k = 0; $k < scalar(@lbl); $k++) {
         $distance = 0;
         for ($l = 5; $l <= 23; $l += 3) {
            $euc_sum = 0;
            for ($j = $l; $j < $l + 3; $j++) {
               $euc_sum += (($lbl[$k][$j] - $dsn[$i][$j - 3]) ** 2);
            }
            $distance += sqrt($euc_sum);
         }
         if (($distance < $min_d || $min_k == -1) && ((! $decision) || ($decision && $distance <= $decision))) {
            $min_k = $k;
            $min_d = $distance; 
         }
      }
      $dtc[$i][0] = "IS_".$lbl_i;
      if ($min_k == -1) {
         $dtc[$i][1] = "AS_ATTACK"; 
      } else {
         if ($lbl[$min_k][4] eq "NORMAL") {
            $dtc[$i][1] = "AS_NORMAL"; 
         } else {
            $dtc[$i][1] = "AS_ATTACK"; 
         }
      }
      if ((($i+1) % 1000) == 0) {
         # view
         print "detect - item(".($i+1).")\n";
      }
   }
   return @dtc;
}
###############################################################################


###############################################################################
sub save_dtc {
   my $file = shift;
   my @dtc = @_;
   my $i;
   open (FILE, "> $file");
   print FILE "req_label cluster_label\n";
   for ($i = 0; $i < scalar(@dtc); $i++) {
      print FILE $dtc[$i][1]." ";
      print FILE $dtc[$i][0]." ";
      print FILE "\n";
   }
   close (FILE);
}
###############################################################################


###############################################################################
sub detection_results {
   my @dtc = @_;
   my ($tp,$fp,$fn,$tn,$ad,$an) = (0,0,0,0,0,0);
   for (my $i = 0; $i < scalar(@dtc); $i++) {
      if ($dtc[$i][0] eq "IS_NORMAL") {
         if ($dtc[$i][1] eq "AS_NORMAL") {
            $tn++;
         } else {
            $fp++;
         }
      } elsif ($dtc[$i][0] eq "IS_ATTACK") {
         if ($dtc[$i][1] eq "AS_ATTACK") {
            $tp++;
         } else {
            $fn++;
         }
      } elsif ($dtc[$i][0] eq "IS_ANOMALY") {
         if ($dtc[$i][1] eq "AS_ATTACK") {
            $ad++;
         } else {
            $an++;
         }
      }
      if ((($i+1) % 10000) == 0) {
         # view
         print "detect - item(".($i+1).") - TP $tp FP $fp FN $fn TN $tn AD $ad AN $an\n";
      }
   }
   return ($tp,$fp,$fn,$tn,$ad,$an);
}
###############################################################################


###############################################################################
sub save_res {
   my $file = shift;
   my @res = @_;
   open (FILE, "> $file");
   print FILE "TP ".$res[0]."\n";  # true positive
   print FILE "FP ".$res[1]."\n";  # false positive
   print FILE "FN ".$res[2]."\n";  # false negative
   print FILE "TN ".$res[3]."\n";  # true negative
   print FILE "AD ".$res[4]."\n";  # anomaly detected
   print FILE "AN ".$res[5]."\n";  # anomaly non-detected
   close (FILE);
}
###############################################################################

