#!/usr/bin/perl

# Article:
# [Pereira and Jamhour, 2011]
# Metodo Heuristico para Rotular Grupos em Sistema de Deteccao de Intrusao baseado em Anomalia

# URL: http://ids.celepar.pr.gov.br/scripts

# Intrusion Detection
# Self-detecting - Zhong (D2)

use strict;

my $ifdie = "Example: perl $0 dts.p01.txt.dsn.100.cls.zhong.0.115.lbl dts.p01.txt.dsn 0.115\nResults: dts.p01.txt.dsn.100.cls.zhong.0.115.lbl.in.dts.p01.txt.dsn.0.115.dtc,\n dts.p01.txt.dsn.100.cls.zhong.0.115.lbl.in.dts.p01.txt.dsn.0.115.res\n\n";
my $file = $ARGV[0] || die "$ifdie";
my $file2 = $ARGV[1] || die "$ifdie";
my $perc = $ARGV[2] || die "$ifdie";

my $file_lbl = $file;
my $file_dsn = $file2;
my $file_dtc = $file.".in.".$file2.".".$perc.".dtc";
my $file_res = $file.".in.".$file2.".".$perc.".res";

my @lbl = &load_labeledclusters($file_lbl);
my @dsn = &load_dsn($file_dsn);

my @dtc = &self_detecting($perc,\@lbl,\@dsn);
&save_dtc($file_dtc,@dtc);
print "done - $file_dtc\n";

my @res = &detection_results(@dtc);
&save_res($file_res,@res);
print "done - $file_res\n";

##############################################################################
sub load_labeledclusters {
   my $file = shift;
   my @lbl;
   my ($k,$line);
   my $fline = 1;
   open (FILE, $file);
   while ($line = <FILE>) {
      if (! $fline) {
         chomp($line);
         @{$lbl[$k]} = split(/ +/,$line);
         $k++;
      } else {
         $fline = 0;
      }
   }
   close (FILE);
   return @lbl;
}
##############################################################################


###############################################################################
sub load_dsn {
   my $file = shift;
   my $fline = 1;
   my @dsn;
   my $c = 0;
   my $line;
   open (FILE, $file);
   while ($line = <FILE>) {
      if (! $fline) {
         chomp ($line);
         @{$dsn[$c]} = split(/ /,$line);
         $c++;
         if ((($c+1) % 10000) == 0) {
            # view
            print "load - item(".($c+1).")\n";
         }
      } else {
         $fline = 0;
      }
   }
   close (FILE);
   return @dsn;
}
###############################################################################


###############################################################################
sub self_detecting {
   my $perc = shift;
   my ($ref_lbl,$ref_dsn) = @_;
   my @lbl = @{$ref_lbl};
   my @dsn = @{$ref_dsn};
   my @dtc;
   my $k_centroid = 0; 
   my $k;
   for ($k = 0; $k < scalar(@lbl); $k++) {
      if ($lbl[$k][3] == 0) {
         $k_centroid = $k;
      }
   }
   my %sort;
   my ($i,$j,$l,$n,$distance,$euc_sum,$sum_attack);
   for (my $i = 0; $i < scalar(@dsn); $i++) {
      $distance = 0;
      for ($l = 5; $l <= 23; $l += 3) {
         $euc_sum = 0;
         for ($j = $l; $j < $l + 3; $j++) {
            $euc_sum += (($lbl[$k_centroid][$j] - $dsn[$i][$j - 3]) ** 2);
         }
         $distance += sqrt($euc_sum);
      }
      if ((($i+1) % 10000) == 0) {
         # view
         print "distance - item(".($i+1).")\n";
      }
      $dtc[$i][0] = "IS_".$dsn[$i][1];
      $sort{$i} = $distance;
   }
   my @sort;
   foreach $i (sort {$sort{$b} <=> $sort{$a}} keys %sort) {
      if ((($n+1) % 10000) == 0) {
         # view
         print "sort - item(".($n+1).")\n";
      }
      $sort[$n] = $i;
      $n++;
   }
   for ($n = 0; $n < scalar(@sort); $n++) {
      $i = $sort[$n];
      if (($sum_attack / scalar(@dsn)) < $perc) {
         $dtc[$i][1] = "AS_ATTACK";
         $sum_attack ++;
      } else {
         $dtc[$i][1] = "AS_NORMAL";
      }	
      if ((($n+1) % 10000) == 0) {
         # view
         print "detect - item(".($n+1).")\n";
      }
   }
   return @dtc;
}
###############################################################################


###############################################################################
sub save_dtc {
   my $file = shift;
   my @dtc = @_;
   my $i;
   open (FILE, "> $file");
   print FILE "req_label cluster_label\n";
   for ($i = 0; $i < scalar(@dtc); $i++) {
      print FILE $dtc[$i][1]." ";
      print FILE $dtc[$i][0]." ";
      print FILE "\n";
   }
   close (FILE);
}
###############################################################################


###############################################################################
sub detection_results {
   my @dtc = @_;
   my ($tp,$fp,$fn,$tn,$ad,$an) = (0,0,0,0,0,0);
   for (my $i = 0; $i < scalar(@dtc); $i++) {
      if ($dtc[$i][0] eq "IS_NORMAL") {
         if ($dtc[$i][1] eq "AS_NORMAL") {
            $tn++;
         } else {
            $fp++;
         }
      } elsif ($dtc[$i][0] eq "IS_ATTACK") {
         if ($dtc[$i][1] eq "AS_ATTACK") {
            $tp++;
         } else {
            $fn++;
         }
      } elsif ($dtc[$i][0] eq "IS_ANOMALY") {
         if ($dtc[$i][1] eq "AS_ATTACK") {
            $ad++;
         } else {
            $an++;
         }
      }
      if ((($i+1) % 10000) == 0) {
         # view
         print "detect - item(".($i+1).") - TP $tp FP $fp FN $fn TN $tn AD $ad AN $an\n";
      }
   }
   return ($tp,$fp,$fn,$tn,$ad,$an);
}
###############################################################################


###############################################################################
sub save_res {
   my $file = shift;
   my @res = @_;
   open (FILE, "> $file");
   print FILE "TP ".$res[0]."\n";  # true positive
   print FILE "FP ".$res[1]."\n";  # false positive
   print FILE "FN ".$res[2]."\n";  # false negative
   print FILE "TN ".$res[3]."\n";  # true negative
   print FILE "AD ".$res[4]."\n";  # anomaly detected
   print FILE "AN ".$res[5]."\n";  # anomaly non-detected
   close (FILE);
}
###############################################################################

